package examples.jce.rsa_aes;

import java.io.*;
import javax.crypto.*;
import javax.crypto.spec.*;
import java.security.*;
import java.security.cert.*;


/**
 * Encrypt a file and use signatures MD5 with RSA for message validation and authentication.
 * This class can be used as a command line tool..
 * EncryptFileSIG [source file] [dest file] [recipients certificate] [your trust store]
 * 
 */



public class EncryptFileSIG
{

    private File file_in = null; // Source File.
    private File file_out = null; // Output File.
    private SecureRandom sec_rand = null; // Secure Random number generator.
    private PublicKey pub_key = null; // Public key.
    private X509Certificate my_cert = null;
    private PrivateKey my_private = null;

 

    public static void main(String[] args)
    {
        new EncryptFileSIG(args);
    }


    /**
     * Read from commans line constructor.
     */

    
    public EncryptFileSIG(String[] args)
    {
        try {
            initTerm(args);
            encryptFileSIG();
        } catch (Exception ex)
            {
                ex.printStackTrace();
            }
    }




    /**
     * Initialise from command line arguments.
     */


    public void initTerm(String[] args)
    { 
        System.out.println("Encrypt file with and use Signature (MD5withRSA) to provide"); // Print statements.
        System.out.println("sender authentication and message integrity checking"); // More print statements.

        sec_rand = null; // Secure random number


        //
        // Add BouncyCastle JCE as a provider and create and seed the random number generator.
        //

        try
        {
            Security.addProvider(new org.bouncycastle.jce.provider.BouncyCastleProvider()); // Add provider.
            sec_rand = SecureRandom.getInstance("SHA1PRNG","SUN"); // Create random number generator.

            //
            // Try and seed PRNG from '/dev/urandom' on UNIX machines this device generally exists.
            // if it does not then SecureRandom will employ its own platform independent methods.
            //

            if (new File("/dev/urandom").exists()) // Does device micro random exist.
            {
                System.out.println("Salting SecureRandom (SHA1PRNG) from /dev/urandom");
                byte[] salt = new byte[8192]; // salt.
                new FileInputStream("/dev/urandom").read(salt); // Read salt.
                sec_rand.setSeed(salt); // Set salt.
            }
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
        }

        //
        // Check from minimum number of command line arguments.
        //

        if (args.length <4)
        {
            System.out.println("Usage: <source file> <dest file> <recipients certificate> <your trust store>");
            System.exit(0);
        }

        file_in = new File(args[0]); // Source of plain text.
        file_out = new File(args[1]); // Destination of data.

        //
        // Check that source file exists.
        //

        if (!file_in.exists())
        {
            System.out.println("Source File does not exist..!");
            System.exit(0);
        }
    
        //
        // Check that recipients certificate file exists.
        // 

        File    cert_file = new File(args[2]); // Certificate file.

        if (!cert_file.exists()) // Does it _not_ exist.
        {
            System.out.println("Recipients Certificate File does not exist..!");
            System.exit(0);
        }

        //
        // Check that key store file exists. 
        //

        File    keystore_file = new File(args[3]);

        if (!keystore_file.exists()) // Does it _not_ exist.
        {
            System.out.println("Keystore file specified could not be found.");
            System.exit(0);
        }



        char[]          ks_pass = null; // Keystore password.
        boolean         stop = false; // Stop flag.
        KeyStore        key_store = null; // Key Store object.

        try
        {
            key_store = KeyStore.getInstance("JKS", "SUN"); // Create a new instance of the JKS key store.        

            //
            // We need to obtain the password for the key store from the user.
            // Use our getLine() method to read it in.
            // Exit program on zero length input.
            // Keep looping until user inputs a password that works.
            //

            while(!stop)
            {
                System.out.println("");
                System.out.print("Your Key Store Password > ");
                String l = getLine(); // Read line.

                if (l == null) // Null entry.
                {
                    System.exit(0);
                }
                
                if (l.length() == 0) // Zero length.
                {
                    System.exit(0);
                }

                ks_pass = l.toCharArray(); // Need to convert to char[] .

                //
                // Test password on keystore.
                // KeyStore will break when the password is incorrect and
                // throw an exception.
                //

                try
                {
                    key_store.load(new FileInputStream(keystore_file),ks_pass);
                    stop=true; // it worked so stop looping..
                }
                catch (Exception ex)
                {
                    System.out.println(ex.getMessage());
                }
            }
        
            stop = false; // Reset stop flag.
        

            //
            // We now need to get an alias for the private key and our certificate 
            // that we will use to sign the encrypted file so that the recipient
            // can authenticate us.
            // This uses the same process as described above.

            while(!stop)
            {
                System.out.println("");
                System.out.print("Secret Key Alias for Signing only use > ");
                String l = getLine();
                if (l == null)
                {
                    System.exit(0);
                }
                
                if (l.length() == 0)
                {
                    System.exit(0);
                }
                
                try
                {
                    my_private = (PrivateKey)key_store.getKey(l, ks_pass); // Get private key.
                    my_cert = (X509Certificate)key_store.getCertificate(l); // Get Signing Certificate
                    stop = true;
                }
                catch (Exception ex)
                {
                    System.out.println(ex.getMessage());
                }
            }


            //
            // Now we need to read in the certificate of the recipient.
            // The certificate contains the public key that the recipient uses for encryption.
            // See javadoc for CertificateFactory for more info..
            //
        
            System.out.println("Reading recipients certificate file.");
      
            byte[]          raw = new byte[(int)cert_file.length()]; // Create new array
            DataInputStream din = new DataInputStream(new FileInputStream(cert_file)); // Some stream wrapper.

            din.readFully(raw); // Read it all in.
            din.close(); // Close file input.
    
            ByteArrayInputStream    b_in = new ByteArrayInputStream(raw); // Need to use this..
            
            //
            // Use a certificate factory to convert the raw certificate into a certificate object.
            //

            CertificateFactory      cf = CertificateFactory.getInstance("X.509"); // Create certificate factory.
            X509Certificate         x_cert = (X509Certificate)cf.generateCertificate(b_in); // Create certificate.
            pub_key = x_cert.getPublicKey();
        } catch (Exception ex)
            {
                ex.printStackTrace();
            }
    }




    /**
     * Encrpyt file.
     */

    public void encryptFileSIG() throws Exception
    {
        EncryptSIG.processFile(file_in, pub_key, my_cert, my_private, sec_rand, file_out);
    }
    

   
  
    /**
     * Read in line from System.in.
     */

    private static String getLine()
    {
        try
        {
            BufferedReader bin = new BufferedReader(new InputStreamReader(System.in));
            return(bin.readLine());
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
        }

        return(null);
    }

} 
