package examples.jce.rsa_aes;

import java.io.*;
import javax.crypto.*;
import javax.crypto.spec.*;
import java.security.*;
import java.security.cert.*;



/**
 * Encrtypt files using HMAC.
 */



public class EncryptHMAC
{
    static final int           FILE_HEADER = 0x7e01;
    static final int           DATA_BLOCK = 1;
    static final int           FINAL_DATA_BLOCK = 2;
    static final int           HMAC_BLOCK = 3;
   
    static final int           KEY_BLOCK = 16;
    static final int           IV_BLOCK = 17;
    static final int           HMAC_KEY_BLOCK = 18;

 
    /**
     * zero out the passed in byte array
     */
    private static void blank(
        byte[]   bytes)
    {
        for (int t = 0; t < bytes.length; t++)
        {
            bytes[t]=0;
        }
    }


   /**
     * This method encrypts the file and forms a header
     * which has components encrypted with RSA using the public
     * key obtained from the certificat of the recipient.
     */

    public static void encryptFile(
                            File            file_in,  // Source file.
                            X509Certificate x_cert, // Recipients certificate
                            SecureRandom    sec_rand, // Secure Random instance
                            File            file_out) // Output File.
        throws Exception
    {
        encryptFile(file_in, x_cert.getPublicKey(), sec_rand, file_out); // Call real method.
    }

    

    /**
     * This method encrypts the file and forms a header
     * which has components encrypted with RSA using the public
     * key of the recipient.
     */

    public static void encryptFile(
        File            file_in,
        PublicKey       pub_key,
        SecureRandom    sec_rand,
        File            file_out)
        throws Exception
    {
        //
        // Generate key for AES cipher at 128 bits.
        //
    
        System.out.println("Generate key for AES at 128 bits");

        KeyGenerator key_gen = KeyGenerator.getInstance("AES", "BC"); // Using appropriate key generator.
        key_gen.init(128, sec_rand); // Set up with key size and a SecureRandom.
        Key aes_key = key_gen.generateKey(); // Generate the Key Object.
    
        //
        // Instantiate Cipher
        //
        System.out.println("Set up cipher AES/CBC/PKCS7Padding");
        Cipher  output_cipher = Cipher.getInstance("AES/CBC/PKCS7Padding","BC"); // Get cipher.

        output_cipher.init(Cipher.ENCRYPT_MODE,aes_key,sec_rand); // initialize with aes_key.
    
        //
        // Get key and IV for cipher so that they can be later encrypted
        // to build a header.
        //
        byte[] aes_iv = output_cipher.getIV();


        //
        // Get encoded form of the AES key.
        // We later store this in the file header encrypted with the public key of the recipient.
        //

        byte[] aes_key_enc = aes_key.getEncoded();

    
        //
        // Set up locking data so that one piece of the Signature info or
        // HMAC is based on data that is not directly transmitted. If this
        // is not done the message could be resigned by an attacker if they
        // also have the recipient's public key.  (Which you should expect.)
        //
        // I generally prefer to use locking data like this so that the
        // message can be validated as original without having to be fully
        // decrypted first.
        //
        // As another approach you could base a Signature or HMAC on the
        // plain text of the message the result is still the same as if you
        // had chosen suitably large locking data.
        //
        // NOTE: In the case of an HMAC the locking data is the key to the HMAC.
        //
    
        //
        // Setup HMAC
        //
        Mac mac = Mac.getInstance("HMACSHA1","BC");
    
        //
        // At the moment (v114) there is not a specific HMACSHA1 key
        // generator and if it did exist it would do the same thing as the
        // use of SecureRandom below.
        //
        System.out.println("Generate key and then set up HMAC (HMACSHA1).");

        //
        // Note also that the size of the HMAC key is recommended to be
        // the same as the output length of the underlying its hash function.
        // So SHA1 has a 160 bit output so therefore the key for HMACSHA1 is 20 bytes.
        //

        byte[]  mac_key_bytes = new byte[20]; // Set to correct length for HMACSHA1
        sec_rand.nextBytes(mac_key_bytes); // Fill with random data,
        
        Key mac_key = new SecretKeySpec(mac_key_bytes, "HMACSHA1"); // Convert to key object.
        mac.init(mac_key); // Initialize.
    
        //
        // Setup RSA to encrypt secrets.
        //
        System.out.println("Set up RSA with OAEPPadding (see PKCS1 V2)");

        Cipher  rsa_eng = Cipher.getInstance("RSA/None/OAEPPadding","BC"); // Get Cipher instance for RSA. 
        
        //
        // Initialize for encryption with public key of recipient.
        //
        
        rsa_eng.init(Cipher.ENCRYPT_MODE,pub_key,sec_rand); 
    
        //
        // Setup to process File.
        //
        FileInputStream     in_str = new FileInputStream(file_in); // Stream to read plain text from.
        
        //
        // Set up output stream wrappers to help us store data and also
        // automatically update the MAC with our output as we are working.
        //

        FileOutputStream    file_str = new FileOutputStream(file_out); // File out. 
        MacOutputStream     mac_str = new MacOutputStream(file_str,mac); // Mac filter.
        DataOutputStream    data_str = new DataOutputStream(mac_str); // Data out.
    
        System.out.println(file_out.getName()+" >> file header");
        data_str.writeShort(FILE_HEADER); // Write a file header.

    
        //
        // Write Key block.
        // This block contains the encoded AES key and is encrypted with RSA.
        //

        System.out.println(file_out.getName()+" >> Aes key enc with RSA");
        data_str.writeShort(KEY_BLOCK); // Write block header.
        byte[] tmp = rsa_eng.doFinal(aes_key_enc); // Encrypt key.
        data_str.writeInt(tmp.length); // Write out length.
        data_str.write(tmp); // Write encrypted key.
        blank(tmp); // Erase array.
    
        //
        // This will upset some people because you don't have to keep the IV
        // a secret. I keep the IV secret because I prefer to keep things
        // secret when I can. 
        // 
        System.out.println(file_out.getName()+" >> Aes IF enc with RSA (See note in source code)");
        data_str.writeShort(IV_BLOCK); // Write block header.
        tmp = rsa_eng.doFinal(aes_iv); // Encrypt IV.
        data_str.writeInt(tmp.length); // Write length.
        data_str.write(tmp); // Write encrypted IV.
        blank(tmp); // Erase array.
    
        //
        // Write out key for HMAC.
        //
        System.out.println(file_out.getName()+" >> HMACSHA1 key enc with RSA");


        //
        // Note we are using the AES cipher to encrypt the HMAC Secret.
        //

        data_str.writeShort(HMAC_KEY_BLOCK);  // Write block header.      
        tmp = output_cipher.doFinal(mac_key.getEncoded()); // Encrypt encoded HMAC key
        data_str.writeInt(tmp.length); // Write length.
        data_str.write(tmp); // Write encrypted key.
        blank(tmp); // Erase array.


        //
        // Reset AES cipher back to original, as you can not rely on the doFinal() method to do it.
        //

            

        //
        // This next block performs the encryption.
        // The plain text is broken up into blocks of 8192 bytes.
        // The blocks are processed through the symmetric cipher.
        // The output is saved in a 'DATA_BLOCK' that is length defined.
        //

        int     l = 0;  // Universal length variable.
        byte[]  buf = new byte[8192]; // Block to process data in.
        byte[]  out = null; // output.

        while((l = in_str.read(buf)) > -1) // Read in data.
        {
            out = output_cipher.update(buf,0,l); // Encrypt (l) bytes of data.
            if (out != null) 
		{
	 		data_str.writeShort(DATA_BLOCK); // Write Block header.
            		data_str.writeInt(out.length); // Write out length of this hunk of cipher text.
            		data_str.write(out); // Write cipher text.
            		System.out.println(file_out.getName()+" >> Encrypted "+out.length+" bytes output");
        	}
	}

        //
        // When the input file has been processed the final data block needs to be
        // saved. This block will contain any final data that was processed that was
        // less than the block size of the cipher padded (PKCS7Padding) to the correct
        // block length. 

        out = output_cipher.doFinal(); // Not call to doFinal() Cipher is reset at this point.
        data_str.writeShort(FINAL_DATA_BLOCK); // Write block header.
        data_str.writeInt(out.length); // Write length.
        data_str.write(out); // Write final cipher text.
        System.out.println(file_out.getName()+" >> Final Encrypted "+out.length+" bytes output");


        blank(buf); // Clear buffer.
        buf = null; // Set to null.

        //
        // The last block in the encrypted file contains the
        // HMAC code. This is the output of the MAC and will be used
        // to verify the integrity of the file during decryption.
        //

        System.out.println(file_out.getName()+" >> Write out HMAC code block.");
        
        
        data_str.writeShort(HMAC_BLOCK); // Write header.
        data_str.flush(); // Flush stream.
        
        //
        // Calculate mac code.
        //
        tmp = mac.doFinal();
        data_str.writeInt(tmp.length); // Write length.
        data_str.write(tmp); // Write code.
        blank(tmp); // Blank array.
    
        data_str.flush(); // Flush output.
        data_str.close(); // Close.
    


        //
        // Get rid of key;
        // Attention needs to be payed to what you do with your secrets.
        // after you have finished using them.
        // Java does not allow good destruction of secrets so we do the
        // best that we can.
        //

        System.out.println("Dispose of key material as best we can in java..");

        blank(aes_key_enc); // blank encoded key.
        aes_key_enc = null; // Set to null;
        aes_key = null; // Best we can do on a Key object; 
        tmp = null; // Get rid of tmp also.
    
        System.out.println("The End..");
    }



}
