package examples.jce.session;

import java.io.*;
import java.net.*;
import java.util.*;

import javax.crypto.*;
import java.security.*;
import java.security.spec.*;
import java.security.cert.*;

import javax.crypto.spec.*;

/**
 * This is the client side of the session exchange.
 * In this session establishment system the client side
 * leads the server.
 * However the server will timeout after 10 seconds if the client
 * pauses.
 */
public class ClientSide
{
    private KeyStore        key_store;
    private char[]          ks_pass;
    private int             port;
    private SecureRandom    sec_rand;

    /**
     * zero out the passed in byte array
     */
    private static void blank(
        byte[]   bytes)
    {
        for (int t = 0; t < bytes.length; t++)
        {
            bytes[t]=0;
        }
    }

    /**
     * Client Side of the session.
     *
     */
    public ClientSide(
        KeyStore        key_store, 
        char[]          ks_pass, 
        int             port, 
        SecureRandom    sec_rand)
    {
        this.key_store = key_store;
        this.ks_pass = ks_pass;
        this.port = port;
        this.sec_rand = sec_rand;
    }

    /**
     * Invoked to start session setup
     *
     */
    public void doClientSide() throws Exception
    {
        byte[]  se_rand = new byte[32]; // Server Random
        byte[]  cl_rand = new byte[32]; // Client Random
        
        Cipher  cl_cipher = Cipher.getInstance("AES/CBC/PKCS7Padding", "BC"); // Client Side cipher.
        Cipher  se_cipher = Cipher.getInstance("AES/CBC/PKCS7Padding", "BC"); // Server Side cipher.
        Cipher  rsa_eng = Cipher.getInstance("RSA/None/OAEPPadding", "BC"); // Cipher created to do RSA with OAEPPadding
        Mac     cl_mac = Mac.getInstance("HMACSHA1", "BC"); // Client Side MAC
        Mac     se_mac = Mac.getInstance("HMACSHA1", "BC"); // Server Side MAC

        //
        // Open socket and set up stream wrappers..
        //
        Socket              sock = new Socket("127.0.0.1", port); // The socket.
    
        //
        // Wrap some special stream wrappers that support PDU generation
        // and a writeByteArray(byte[] b) method for us to use.
        //

        ExDataOutputStream  data_out = new ExDataOutputStream(sock.getOutputStream());
        ExDataInputStream   data_in = new ExDataInputStream(sock.getInputStream());
        
        //
        // Step 1.
        //
        // Send a HELLO to server + my random value.
        //
        
        data_out.writeShort(SessionConstants.HELLO);
        sec_rand.nextBytes(cl_rand);
        data_out.write(cl_rand);
        data_out.flush();

        //
        // The server will now respond with its Hello + random.
        //
        // Read the HELLO back from the server and collect
        // the Server Random value.
        //

        if (data_in.readShort() == SessionConstants.HELLO)
        {
            // 
            // Read back server random.
            //
            data_in.readFully(se_rand);
        }
        else
        {
            throw (new Exception("Session broken at HELLO stage"));
        }


        //
        // Next send my public key from the key pair that is only
        // used for encryption/decryption purposes. Then sign the whole
        // exchange with my signing only key pair.
        //

        //
        // Set up Signature Class.
        //

        Signature   sig = Signature.getInstance("MD5withRSA", "BC");
    
        //
        // Init signature with the private key used for signing.
        //

        sig.initSign((PrivateKey)key_store.getKey("clientsig", ks_pass), sec_rand);

        //
        // All signatures incorporate the client random and the server
        // random values.
        //
        sig.update(cl_rand);    // Incorporated into every sig.
        sig.update(se_rand); 
        
        //
        // Get my public key (for encryption) as a byte array.
        //
        byte[]  my_pub = ((X509Certificate)key_store.getCertificate("clientenc")).getPublicKey().getEncoded();

        //
        // Get my certificate (for sig validation and auth) as a byte array.
        //
        byte[] my_cert = ((X509Certificate)key_store.getCertificate("clientsig")).getEncoded();
        
        sig.update(my_pub);     // Incorporate public key into signature.
        sig.update(my_cert);    // Incorporate certificate into signature.
    
        //
        // Write out header to indicate Public key exchange                     
        //
        data_out.writeShort(SessionConstants.PUBLIC_KEY_EXCHANGE);
        
        //
        // Write out my public key.
        //
        data_out.writeByteArray(my_pub);
        
        //
        // Write out certificate.
        //
        data_out.writeByteArray(my_cert);

        //
        // Generate a signature code by calling sig.sign();
        //
        byte[]  sig_code = sig.sign();
    
        //
        // Write this as the last part of the PDU and flush the stream.
        //
        data_out.writeByteArray(sig_code);
        data_out.flush();
        
        System.out.println("Sent to server");

        PublicKey se_public = null;
        X509Certificate server_cert = null;

        //
        // Now server should respond with its public key exchange message.
        // If it does not I must break as the protocol has been broken.
        //
        if (data_in.readShort() == SessionConstants.PUBLIC_KEY_EXCHANGE)
        {
            System.out.println("Client: Reading public key + cert + sig from server");

            //
            // Read in Server Public key.
            //
            byte[]              pub_key = data_in.readByteArray();
            
            //
            // Before we can use the public key we must convert it back
            // to a Key object by using the KeyFactory and the appropriate
            // KeySpec.. In this case the X509EncodedKeySpec is the correct one.
            //

            X509EncodedKeySpec  key_spec = new X509EncodedKeySpec(pub_key);
            KeyFactory          key_fact = KeyFactory.getInstance("RSA","BC");

            //
            // Recover Servers Public key.
            //
            se_public = key_fact.generatePublic(key_spec);

            //
            // Read in the encoded form of the X509Certificate that the
            // server uses. For authentication of its identity.
            //
            byte[]              cert = data_in.readByteArray();

            //
            // Set up a Certificate Factory to process the raw certificate
            // back into an X509 Certificate
            //
            CertificateFactory  cf = CertificateFactory.getInstance("X.509");

            //
            // Recover Servers Certificate
            //
            server_cert = (X509Certificate)cf.generateCertificate(new ByteArrayInputStream(cert));
            
            // NOTE:
            // At this point it should be noted that the client must employ
            // some mechanism to validate and authenticate the servers
            // certificate.

            //
            // Read in Signature code.
            //
            sig_code = data_in.readByteArray();
            
            //
            // Now we must verify the data that we received against
            // the signature code at the end of the pdu.
            //

            //
            // Using the authentication certificate sent by the server.
            //
            sig.initVerify(server_cert);
            sig.update(cl_rand);    // Incorporate in Client Random.
            sig.update(se_rand);    // Incorporate in Server Random.
            sig.update(pub_key);    // Incorporate in Public key (as sent in encoded form).
            sig.update(cert);       // Incorporate in Certificate. (as sent in encoded form).
            
            if (!sig.verify(sig_code))
            {
                throw new Exception("(CLIENT)Signature failed on Public key exchange data unit");
            }
        }
        else
        {
            throw new Exception("Session broken at PUBLIC_KEY_EXCHANGE stage");
        }
    
        //
        // Now that we have successfully exchanged public keys
        // The client now needs to being a SecretKey Exchange process.
        // First we need to generate some secrets using the appropriate
        // KeyGenerator. When using the JCE you should always use a
        // KeyGenerator instance specifically set up for your target cipher.
        // The KeyGenerator code will generate a secret that is "safe" for use
        // and filter out any keys that may be weak or broken for that cipher.
        //

        KeyGenerator    key_gen = KeyGenerator.getInstance("AES","BC"); // Get instance for AES

        key_gen.init(192, sec_rand); // Use a 192 bit key size.

        SecretKey       my_aes_key = key_gen.generateKey(); // Generate the  SecretKey
        
        key_gen.init(160, sec_rand); // Set up for a 160 bit key.
    
        //
        // Note: At the moment there is no specific KeyGenerator for HMACSHA1.
        // A secret for an MAC of this type is recommended to be the same length
        // as the underlying message digest. (SHA1 is 160 bits).
        // Therefore we need to generate a secret that is 160 bits long.
        //
        SecretKey       my_hmac_key = key_gen.generateKey(); // Generate key for HMAC.
        
        //
        // For AES in CBC mode requires an IV equal in length to the block
        // size of the cipher. AES has a bock size of 128 bits so the IV is a
        // constant block size of 128 bits regardless of key size.
        //
        byte[]          my_iv = new byte[16];

        //
        // Next we need it instantiate the Cipher class that will be used by
        // the client. As this is the client side the cipher needs to be set
        // up for Encryption.
        //
        cl_cipher.init(Cipher.ENCRYPT_MODE, my_aes_key, new IvParameterSpec(my_iv), sec_rand);
        
        //
        // Set up Client side MAC with key.
        //
        cl_mac.init(my_hmac_key);
        
        //
        // I have added this extra piece of data so that an attacker
        // cannot resign or modify the secret exchange without being one of the recipients.
        //
        byte[]          tmp_lock = new byte[24];
    
        //
        // Load byte array with random data.
        //
        sec_rand.nextBytes(tmp_lock); 
        
        //
        // Set up RSA for encryption.
        //
        rsa_eng.init(Cipher.ENCRYPT_MODE, se_public, sec_rand);
        
        //
        // Set up and instace of AES for the Signature locking data.
        //
        
        Cipher aes_lock = Cipher.getInstance("AES/CBC/PKCS7Padding");
        aes_lock.init(Cipher.ENCRYPT_MODE, my_aes_key, new IvParameterSpec(my_iv), sec_rand);


        //
        // Write out marker for SecretExchange.
        //
        data_out.writeShort(SessionConstants.SECRET_EXCHANGE);

        //
        // Set up Signature so that the server can validate our data.
        //
        sig.initSign((PrivateKey) key_store.getKey("clientsig",ks_pass));
        sig.update(cl_rand); // Incorporate Client Random
        sig.update(se_rand); // Incorporate Server Random.
        
        byte[] tmp = rsa_eng.doFinal(my_aes_key.getEncoded());  // Encrypt the encoded AES key.
        data_out.writeByteArray(tmp);                           // Send to server.
        sig.update(tmp);                                        // Incorporate into signature.
        blank(tmp);                                             // Erase array with Zero bytes.
        
        tmp = rsa_eng.doFinal(my_iv);                           // Encrypt the IV. 
        data_out.writeByteArray(tmp);                           // Send to server.
        sig.update(tmp);                                        // Incorporate into signature.
        blank(tmp);                                             // Erase array with Zero bytes.
        
        tmp = rsa_eng.doFinal(my_hmac_key.getEncoded());        // Encrypt and encode MAC key.
        data_out.writeByteArray(tmp);                           // Send to server.
        sig.update(tmp);                                        // Incorporate into signature.
        blank(tmp);                                             // Erase array with Zero bytes.
        
        tmp = aes_lock.doFinal(tmp_lock);                        // Encrypt lock data.
        data_out.writeByteArray(tmp);                           // Send lock data.
        sig.update(tmp_lock);                                   // Incorporate plain text of lock data into signature.
        blank(tmp);                                             // Erase.
        
        data_out.writeByteArray(sig.sign());                    // Write out signature and flush stream.
        data_out.flush();

        //
        // Read back the server secret.
        //
        // The server should respond with its secret if it does not then
        // the protocol is broken..
        //
        if (data_in.readShort() == SessionConstants.SECRET_EXCHANGE)
        {
            //
            // Set up RSA in decrypt mode so that we can decrypt their
            // server's secrets.
            //
            rsa_eng.init(
                     Cipher.DECRYPT_MODE,
                     (PrivateKey)key_store.getKey("clientenc",ks_pass),
                     sec_rand);
            
            //
            // Read in secret key.
            //
            byte[]  aes_key_enc = data_in.readByteArray();
            
            //
            // Read in IV
            //
            byte[]  iv_enc = data_in.readByteArray();
            
            //
            // Read in HMAC key.
            //
            byte[]  hmac_key_enc = data_in.readByteArray();
            
            //
            // Read in lock
            //
            tmp_lock = data_in.readByteArray();    
            
            //
            // Now we must validate the received data.
            // NOTE: the need to decrypt the tmp_lock data first.
            //

            //
            // Set up AES lock so we can decrypt lock data.
            //

             aes_lock.init(Cipher.DECRYPT_MODE,
                           (SecretKey)new SecretKeySpec(rsa_eng.doFinal(aes_key_enc), "AES"),
                           new IvParameterSpec(rsa_eng.doFinal(iv_enc)));



            sig.initVerify(server_cert); // Set up using server's certificate.
            
            sig.update(cl_rand); 
            sig.update(se_rand);
            sig.update(aes_key_enc);
            sig.update(iv_enc);
            sig.update(hmac_key_enc);
            sig.update(aes_lock.doFinal(tmp_lock));

            if (!sig.verify(data_in.readByteArray()))
            {
                throw new Exception("Signature failed on Public key exchange data unit");
            }
            else
            {
                System.out.println("Client: Server PDU for secret key exchange signature passed");
            }
            
            //
            // At this point we have successfully exchanged secrets..
            // So now we need to set up a Cipher class to allow us to
            // Decrypt data sent from the server to the client.
            //
            se_cipher.init(Cipher.DECRYPT_MODE,
                       (SecretKey)new SecretKeySpec(rsa_eng.doFinal(aes_key_enc),"AES"),
                       new IvParameterSpec(rsa_eng.doFinal(iv_enc)));
            
            //
            // We also need to set up the MAC so that we can validate
            // data sent from the server.
            //
            se_mac.init((SecretKey)new SecretKeySpec(rsa_eng.doFinal(hmac_key_enc),"AES"));

            // 
            // Set up streams to generate appropriate PDUs.
            // Take a look at ExDataInput and ExDataOutput stream classes.
            // To see this shape of the PDUs that are used to exchange
            // encrypted data.
            // The writePDU(byte[]) method accepts plain text.
            // The byte[] readPDU() method returns plain text.
            //
       
            data_out.initPDU(cl_cipher,cl_mac); // Setup for sending to server.
            data_in.initPDU(se_cipher,se_mac); // Setup for receiving.

            // 
            // Lets use our newly set up exchange.
            // Send something to server.
            //
            data_out.writePDU("Hello Server how are you today...".getBytes("ISO-8859-1"));

            //
            // Lets get something back from the server.
            //
            System.out.println("Server -> Client: " + new String(data_in.readPDU(),"ISO-8859-1"));
 
            //
            // And again to see if the chaining of the blocks is working..
            //

            data_out.writePDU("10.9.8.7.6.5.4.3.2.1".getBytes("ISO-8859-1"));

            // and the servers response.
            System.out.println("Server -> Client: " + new String(data_in.readPDU(),"ISO-8859-1"));

        

       }
        else
        {
            throw new Exception("Session broken at SECRET_EXCHANGE stage");
        }
        sock.close();
    }
}
 
